require 'json'

# {include:file:README.rdoc}
module LsColors
  # Current software version.
  VERSION = '1.0.2'

  # Date of the current version.
  DATE = '2014-10-30'

  # A short description.
  ABOUT = 'Manipulate LS_COLORS settings.'

  # Methods and data for manipulating terminal colors.
  module ColorManipulation
    # The string that begins a terminal-compatible color definition.
    COLOR_HEAD = "\x1b["

    # The prefix for the foreground color (256 colors).
    COLOR_FG = '38;5;'

    # The prefix for the background color (256 colors).
    COLOR_BG = '48;5;'

    # The string that ends a terminal-compatible color definition.
    COLOR_TAIL = 'm'

    # This string resets color and style to default.
    COLOR_RESET = "\x1b[0m"

    # The style codes recognized by the terminal.
    STYLES = {
      bold: '01',
      under: '04',
      blink: '05',
      reverse: '07',
      conceal: '08'
    }

    # Explanation of system-related file codes, for previewing and learning
    # purposes.
    CODE_DESCRIPTIONS = {
      di: 'directory',
      fi: 'file',
      ln: 'symlink',
      pi: 'fifo',
      so: 'socket',
      bd: 'block',
      or: 'orphan',
      mi: 'missing',
      ex: 'executable'
    }

    # The fallback color sheet.
    DEFAULTS = {
      system: [
        ['di', 33, :bold], ['ln', 45], ['mh', 15, 5],
        ['pi', 11, 1], ['so', 13], ['do', 5], ['bd', 11, 232], ['cd', 3, 232],
        ['or', 45, 9], ['mi', 15, 232, :blink], ['su', 15, 196],
        ['sg', 16, 11], ['ca', 226, 196], ['tw', 16, 10], ['ow', 21, 10],
        ['st', 15, 21], ['ex', 35, :bold]
      ],
      extensions: [
        ['sh pl py swf bat exe', 35],
        ['rb rbw rdoc gemspec gem', 197],
        ['png jpg jpeg gif bmp tiff xcf', 111],
        ['mp4 mkv avi ogm mov mpeg flv wmv', 171],
        ['mp3 mp2 ogg flac wav wma ape mid midi', 168],
        ['gb gba nds 3ds bin iso dat grp pak', 202],
        ['rar zip tar gz tgz 7z bz2', 215],
        ['txt odt doc docx pdf djvu', 247],
        ['bak old log', 241]
      ]
    }

    # Test the 16 system colors.
    #
    # @return [String]
    def test_system
      [(0..7), (8..15)].map { |range| test_array range.to_a }.join
    end

    # Test the 6x6x6 color cube.
    #
    # @return [String]
    def test_cube
      output = (0..17).map do |row|
        column = 16 + (row * 6) + (36 * (row / 6))
        cube_1 = (column..column + 5).to_a
        cube_2 = (column + 36..column + 41).to_a
        test_array(cube_1 + cube_2)
      end

      output.join
    end

    # Test the 24 grayscale colors.
    #
    # @return [String]
    def test_grayscale
      [(232..243), (244..255)].map { |range| test_array range.to_a }.join
    end

    # Test an array of color numbers.
    #
    # @param [Array] array the color numbers to test.
    # @return [String]
    def test_array(array)
      output = array.map { |color| color_card(color) } << "\n"

      output.join
    end

    # Returns a preview of a single color.
    #
    # @param [Integer] color
    # @return [String]
    def color_card(color)
      "#{colorizer ['foo', 0, color]}   #{format '%03d', color}#{COLOR_RESET}"
    end

    # Take an array as per the {DEFAULTS} format and create a terminal style
    # escape string.
    #
    # @param [Array] array
    # @return [String]
    def colorizer(array)
      fail "No foreground color in #{array}." unless array[1].is_a? Integer
      "#{COLOR_HEAD}#{color_setting(array)}#{COLOR_TAIL}"
    end

    # Take an array as per the {DEFAULTS} format and create a color definition
    # string.
    #
    # @param [Array] array
    # @return [String]
    def color_setting(array)
      string = ''

      STYLES.each_pair do |style, code|
        next unless array.include?(style) || array.include?(style.to_s)
        string << "#{code};"
      end
      string << "#{COLOR_FG}#{array[1]}"
      string << ";#{COLOR_BG}#{array[2]}" if array[2].is_a? Integer

      string
    end

    # Return a colorized preview of the system file color settings.
    #
    # @return [String]
    def preview_system
      array = @colors[:system].map do |c|
        code = CODE_DESCRIPTIONS.fetch(c.first.to_sym, c.first)
        "#{colorizer(c)}#{code}#{COLOR_RESET}"
      end

      array.join ' '
    end

    # Return a colorized preview of the extension color settings.
    #
    # @return [String]
    def preview_extensions
      string = ''

      @colors[:extensions].each do |c|
        c.first.split.sort.each do |ext|
          string << "#{colorizer(c)}*.#{ext}#{COLOR_RESET} "
        end
        string << "\n"
      end

      string
    end
  end

  # Load, modify, and output configuration.
  module Colors
    extend ColorManipulation

    # The default configuration file. This will be loaded at startup if present.
    RC_FILE = Dir.home << '/.lscolors_ruby.rc'

    class << self
      # Loads or initializes settings. The parameter can be either a Hash or a
      # String (in the latter case it will be interpreted as the file name of
      # a JSON file). If missing, the method will look for a JSON file specified
      # by {RC_FILE}, and if absent, the {DEFAULTS} will be loaded.
      #
      # @param [String, Hash] object
      # @return [Hash] the parsed settings.
      def load(object = nil)
        return @colors = object if object.is_a? Hash

        if object.is_a? String
          return @colors = JSON.parse(File.read(object), symbolize_names: true)
        elsif File.exist? RC_FILE
          return @colors = JSON.parse(File.read(RC_FILE), symbolize_names: true)
        end

        @colors = DEFAULTS
      end

      # Will preview the current settings as a string of colorized system files
      # and extensions. Use +print+ or +puts+ to see the effect (a 256-color
      # terminal is required).
      #
      # @return [String]
      def preview
        "#{preview_system}\n#{preview_extensions}"
      end

      # Save the current settings to a file. Without parameters, {RC_FILE} will
      # be created and/or saved. {RC_FILE} will be loaded automatically from
      # now on, so it can be modified directly in order to configure the color
      # settings.
      #
      # @param [String] file the file name to save to.
      # @return [0]
      def save(file = RC_FILE)
        File.write(file, JSON.pretty_generate(@colors))
        0
      end

      # Test the terminal color capability.
      #
      # @return [String]
      def test
        colors = `tput colors`.to_i
        fail "256 colors not supported! (#{colors})" if colors < 256

        "System colors (0-15):\n#{test_system}\n" \
          "Color cube (16-231):\n#{test_cube}\n" \
          "Grayscale colors (232-255):\n#{test_grayscale}\n"
      end

      # Will output the +LS_COLORS+ compatible configuration string.
      #
      # @return [String]
      def to_s
        output = 'rs=0:'

        @colors[:system].each do |c|
          output << "#{c.first}=#{color_setting(c)}:"
        end

        @colors[:extensions].each do |c|
          c.first.split.sort.each do |ext|
            output << "*.#{ext}=#{color_setting(c)}:"
          end
        end

        output
      end
      alias_method :inspect, :to_s
    end
  end
end

LsColors::Colors.load
